/*
 * Copyright (c) 2008-2016 Haulmont.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package com.haulmont.cuba.client;

import com.haulmont.cuba.core.config.Config;
import com.haulmont.cuba.core.config.Property;
import com.haulmont.cuba.core.config.Source;
import com.haulmont.cuba.core.config.SourceType;
import com.haulmont.cuba.core.config.defaults.*;
import com.haulmont.cuba.core.config.type.CommaSeparatedStringListTypeFactory;
import com.haulmont.cuba.core.config.type.Factory;

import java.util.List;
import java.util.regex.Pattern;

/**
 * Configuration parameters interface used by the WEB and DESKTOP layers.
 */
@Source(type = SourceType.APP)
public interface ClientConfig extends Config {

    /**
     * @return middleware connection URL list
     */
    @Property("cuba.connectionUrlList")
    @Factory(factory = CommaSeparatedStringListTypeFactory.class)
    List<String> getConnectionUrlList();

    /**
     * @return Context of the middleware file download controller.
     */
    @Property("cuba.fileDownloadContext")
    @DefaultString("/download")
    String getFileDownloadContext();

    /**
     * @return Context of the middleware file upload controller.
     */
    @Property("cuba.fileUploadContext")
    @DefaultString("/upload")
    String getFileUploadContext();

    /**
     * @return Maximum size of uploaded file in megabytes.
     */
    @Property("cuba.maxUploadSizeMb")
    @Source(type = SourceType.DATABASE)
    @DefaultInt(20)
    int getMaxUploadSizeMb();
    void setMaxUploadSizeMb(int value);

    @Factory(factory = UniqueConstraintViolationPatternFactory.class)
    @Source(type = SourceType.DATABASE)
    @Property("cuba.uniqueConstraintViolationPattern")
    Pattern getUniqueConstraintViolationPattern();

    /**
     * @return Whether to enable sorting of datasource data on DB (using separate SELECT with ORDER BY clause).
     */
    @Property("cuba.collectionDatasourceDbSortEnabled")
    @DefaultBoolean(true)
    boolean getCollectionDatasourceDbSortEnabled();

    /**
     * @return If true, client will try to find missing localized messages on the server.
     */
    @Property("cuba.remoteMessagesSearchEnabled")
    @DefaultBoolean(false)
    boolean getRemoteMessagesSearchEnabled();

    /**
     * List of screen aliases for which saving screen history is enabled.
     * <p>Obsolete. Recommended way to specify this information is entity annotations
     * in {@code *-metadata.xml}</p>
     * @return  comma-separated list of screen aliases
     */
    @Property("cuba.screenIdsToSaveHistory")
    String getScreenIdsToSaveHistory();

    /**
     * @return Whether to enable password policy. If true, all new passwords will be checked for the compliance with
     * {@code cuba.passwordPolicyRegExp}
     */
    @Property("cuba.passwordPolicyEnabled")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(false)
    boolean getPasswordPolicyEnabled();

    /**
     * @return The regular expression which is used by password policy (see also {@code cuba.passwordPolicyEnabled}).
     */
    @Property("cuba.passwordPolicyRegExp")
    @Source(type = SourceType.DATABASE)
    @DefaultString("((?=.*\\d)(?=.*\\p{javaLowerCase})(?=.*\\p{javaUpperCase}).{6,20})")
    String getPasswordPolicyRegExp();

    /**
     * @return If true, all generic filters will require explicit Apply after screen opening. Empty connected table
     * is shown.<br>
     * If false, the filter will be applied automatically, refreshing the table immediately after the screen opening.
     */
    @Property("cuba.gui.genericFilterManualApplyRequired")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(false)
    boolean getGenericFilterManualApplyRequired();

    /**
     * @return If true, then check filter conditions (empty or not) before applying filter. If all conditions are empty
     * (no parameters entered), the filter doesn't apply and special message to user is shown.<br>
     * If false, no checks are performed and the filter applies, refreshing connected table.
     */
    @Property("cuba.gui.genericFilterChecking")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(false)
    boolean getGenericFilterChecking();

    /**
     * @return a number of nested properties in the "Add Condition" dialog window. I.e. if the depth is 2, then you'll be
     * able to select a property "contractor.city.country", if the value is 3, then "contractor.city.country.name", etc.
     */
    @Property("cuba.gui.genericFilterPropertiesHierarchyDepth")
    @Source(type = SourceType.DATABASE)
    @DefaultInt(2)
    int getGenericFilterPropertiesHierarchyDepth();

    /**
     * @return If true, all generic text filters will trim value.<br>
     * If false, the text filter will not be trim value.
     */
    @Property("cuba.gui.genericFilterTrimParamValues")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getGenericFilterTrimParamValues();

    /**
     * @return Number of columns with conditions in generic filter UI component.
     */
    @Property("cuba.gui.genericFilterColumnsCount")
    @Source(type = SourceType.DATABASE)
    @DefaultInt(3)
    int getGenericFilterColumnsCount();

    /**
     * @return location of panel with conditions in generic filter component.
     * If {@code top} then conditions will be placed above filter control elements
     * or below them if {@code bottom}.
     */
    @Property("cuba.gui.genericFilterConditionsLocation")
    @Source(type = SourceType.DATABASE)
    @Default("top")
    String getGenericFilterConditionsLocation();

    /**
     * Returns a number of items to be displayed in popup list near the 'Search' button. If number of filter
     * entities exceeds this value then 'Show more..' action is added to the popup list. The action
     * will show new dialog window with all possible filter entities for selecting a desired one.
     */
    @Property("cuba.gui.genericFilterPopupListSize")
    @Source(type = SourceType.DATABASE)
    @DefaultInt(10)
    int getGenericFilterPopupListSize();

    /**
     * Returns a template for filter controls layout. Each component has the following format:
     * [<i>component_name</i> | <i>options-comma-separated</i>], e.g. [pin | no-caption, no-icon].
     * <p>Available component names:</p>
     * <ul>
     *     <li>{@code filters_popup} - popup button for selecting a filter entity combined with Search button.
     *     When using this component there is no need to add a separate Search button</li>
     *     <li>{@code filters_lookup} - lookup field for selecting a filter entity. Search button should be added as
     *     a separate component</li>
     *     <li>{@code search} - search button. Do not add if use {@code filters_popup}</li>
     *     <li>{@code add_condition} - button for adding a new condition</li>
     *     <li>{@code spacer} - space between component groups </li>
     *     <li>{@code settings} - settings button. Specify actions names that should be displayed in Settings popup
     *     as component options </li>
     *     <li>{@code max_results} - group of components for setting max number of records to be displayed</li>
     *     <li>{@code fts_switch} - checkbox for switching to FTS mode</li>
     * </ul>
     * The following components can be used as options for {@code settings} component. They also can be used as
     * independent components if for example you want to display a Pin button:
     * <ul>
     *     <li>{@code save}</li>
     *     <li>{@code save_as}</li>
     *     <li>{@code edit}</li>
     *     <li>{@code remove}</li>
     *     <li>{@code pin}</li>
     *     <li>{@code make_default}</li>
     *     <li>{@code save_search_folder}</li>
     *     <li>{@code save_app_folder}</li>
     * </ul>
     * Action components can have the following options:
     * <ul>
     *     <li>{@code no-icon} - if an icon shouldn't be displayed on action button. For example: [save | no-icon]</li>
     *     <li>{@code no-caption} - if a caption shouldn't be displayed on action button. For example: [pin | no-caption]</li>
     * </ul>
     * @return a template for filter controls layout
     */
    @Property("cuba.gui.genericFilterControlsLayout")
    @Source(type = SourceType.DATABASE)
    @Default("[filters_popup] [add_condition] [spacer] [settings | save, save_with_values, save_as, edit, remove, make_default, pin, save_search_folder, save_app_folder, clear_values] [max_results] [fts_switch]")
    String getGenericFilterControlsLayout();

    /**
     * Returns a comma-separated list of values that are used as options for "Show rows" lookup field
     * of generic filter component. Add NULL option to the list if the lookup field should contain an empty value.
     */
    @Property("cuba.gui.genericFilterMaxResultsOptions")
    @Source(type = SourceType.DATABASE)
    @Default("NULL, 20, 50, 100, 500, 1000, 5000")
    String getGenericFilterMaxResultsOptions();

    /**
     * If true, the filter will be applied immediately after every parameter's value changing. True by default.
     *
     * @return true if the filter should be applied immediately after every parameter's value changing
     */
    @Property("cuba.gui.genericFilterApplyImmediately")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getGenericFilterApplyImmediately();

    /**
     * If true, then a context action "Full-Text Search Details" will be added to the table or data grid after full-text search in generic filter
     * component.
     */
    @Property("cuba.gui.genericFilterFtsDetailsActionEnabled")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getGenericFilterFtsDetailsActionEnabled();

    /**
     * If true, then table rows tooltips with hit information will be generated in case of full text search in generic filter component. Tooltips
     * generation may take a lot of time, so be careful when setting the parameter value to "true"
     */
    @Property("cuba.gui.genericFilterFtsTableTooltipsEnabled")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(false)
    boolean getGenericFilterFtsTableTooltipsEnabled();

    /**
     * Support e-mail. Exception report emails are sent to this address.
     */
    @Property("cuba.supportEmail")
    @Source(type = SourceType.DATABASE)
    String getSupportEmail();

    /**
     * @return System ID. Use for identification (support emails).
     */
    @Property("cuba.systemId")
    @DefaultString("CUBA")
    @Source(type = SourceType.DATABASE)
    String getSystemID();

    @Property("cuba.gui.tableShortcut.insert")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-BACKSLASH")
    String getTableInsertShortcut();

    @Property("cuba.gui.tableShortcut.add")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-ALT-BACKSLASH")
    String getTableAddShortcut();

    @Property("cuba.gui.tableShortcut.remove")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-DELETE")
    String getTableRemoveShortcut();

    @Property("cuba.gui.tableShortcut.edit")
    @Source(type = SourceType.DATABASE)
    @DefaultString("ENTER")
    String getTableEditShortcut();

    @Property("cuba.gui.tableShortcut.view")
    @Source(type = SourceType.DATABASE)
    @DefaultString("ENTER")
    String getTableViewShortcut();

    @Property("cuba.gui.commitShortcut")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-ENTER")
    String getCommitShortcut();

    @Property("cuba.gui.closeShortcut")
    @Source(type = SourceType.DATABASE)
    @DefaultString("ESCAPE")
    String getCloseShortcut();

    @Property("cuba.gui.filterApplyShortcut")
    @Source(type = SourceType.DATABASE)
    @DefaultString("SHIFT-ENTER")
    String getFilterApplyShortcut();

    @Property("cuba.gui.filterSelectShortcut")
    @Source(type = SourceType.DATABASE)
    @DefaultString("SHIFT-BACKSPACE")
    String getFilterSelectShortcut();

    @Property("cuba.gui.nextTabShortcut")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-SHIFT-PAGE_DOWN")
    String getNextTabShortcut();

    @Property("cuba.gui.previousTabShortcut")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-SHIFT-PAGE_UP")
    String getPreviousTabShortcut();

    @Property("cuba.gui.pickerShortcut.modifiers")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-ALT")
    String getPickerShortcutModifiers();

    @Property("cuba.gui.pickerShortcut.lookup")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-ALT-L")
    String getPickerLookupShortcut();

    @Property("cuba.gui.pickerShortcut.open")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-ALT-O")
    String getPickerOpenShortcut();

    @Property("cuba.gui.pickerShortcut.clear")
    @Source(type = SourceType.DATABASE)
    @DefaultString("CTRL-ALT-C")
    String getPickerClearShortcut();

    @Property("cuba.gui.useSaveConfirmation")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getUseSaveConfirmation();

    @Property("cuba.gui.loadObsoleteSettingsForTable")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(false)
    boolean getLoadObsoleteSettingsForTable();

    @Property("cuba.gui.layoutAnalyzerEnabled")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getLayoutAnalyzerEnabled();

    @Property("cuba.gui.systemInfoScriptsEnabled")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getSystemInfoScriptsEnabled();
    void setSystemInfoScriptsEnabled(boolean enabled);

    @Property("cuba.gui.manualScreenSettingsSaving")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(false)
    boolean getManualScreenSettingsSaving();

    @Property("cuba.gui.lookupFieldPageLength")
    @Source(type = SourceType.DATABASE)
    @DefaultInt(10)
    int getLookupFieldPageLength();
    void setLookupFieldPageLength(int pageLength);

    /**
     * Interval for checking timeout of a BackgroundTask.
     *
     * @return Timeout in ms
     */
    @Property("cuba.backgroundWorker.timeoutCheckInterval")
    @DefaultInteger(5000)
    Integer getBackgroundTaskTimeoutCheckInterval();

    /**
     * @return true if Table will ignore unfetched attribute read errors
     */
    @Property("cuba.gui.ignoreUnfetchedAttributesInTable")
    @DefaultBoolean(false)
    boolean getIgnoreUnfetchedAttributesInTable();

    @Property("cuba.gui.dynamicAttributesTableColumnMaxTextLength")
    @Source(type = SourceType.DATABASE)
    @DefaultInt(50)
    int getDynamicAttributesTableColumnMaxTextLength();
    void setDynamicAttributesTableColumnMaxTextLength(int maxTextLength);

    /**
     * Defines default behaviour of {@code CreateAction}: whether it will add a new item to the beginning or to the
     * end of datasource's collection. Affects only standalone datasources, for nested datasources new items are always
     * added to the end.
     */
    @Property("cuba.gui.createActionAddsFirst")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getCreateActionAddsFirst();

    @Property("cuba.gui.suggestionField.asyncSearchDelayMs")
    @DefaultInt(300)
    int getSuggestionFieldAsyncSearchDelayMs();

    /**
     * Standard Window validation error notification type.
     *
     * @return one of Notifications.NotificationType values
     */
    @Property("cuba.gui.validationNotificationType")
    @Default("TRAY")
    String getValidationNotificationType();

    /**
     * @return threshold in bytes, if exported byte arrays of {@code ByteArrayDataProvider} bigger than threshold then
     * they will be saved to temp storage to prevent HTTP session memory leaks. Default is 100 KB.
     */
    @Property("cuba.gui.saveExportedByteArrayDataThresholdBytes")
    @DefaultInt(100 * 1024)
    int getSaveExportedByteArrayDataThresholdBytes();

    /**
     * If true, then LookupBuilder and TokenList will reload entity after it is selected from lookup window if the selected entity doesn't contain all
     * required entity attributes
     */
    @Property("cuba.gui.reloadUnfetchedAttributesFromLookupScreens")
    @Source(type = SourceType.DATABASE)
    @DefaultBoolean(true)
    boolean getReloadUnfetchedAttributesFromLookupScreens();
}